/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.operator.keycloak.user;

import java.time.Duration;
import java.util.logging.Level;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import de.ozgcloud.operator.Config;
import de.ozgcloud.operator.keycloak.OzgCustomResourceStatus;
import io.javaoperatorsdk.operator.api.reconciler.Cleaner;
import io.javaoperatorsdk.operator.api.reconciler.Context;
import io.javaoperatorsdk.operator.api.reconciler.ControllerConfiguration;
import io.javaoperatorsdk.operator.api.reconciler.DeleteControl;
import io.javaoperatorsdk.operator.api.reconciler.Reconciler;
import io.javaoperatorsdk.operator.api.reconciler.UpdateControl;
import lombok.extern.java.Log;

@Log
@ControllerConfiguration
@Component
public class KeycloakUserReconciler implements Reconciler<OzgKeycloakUser>, Cleaner<OzgKeycloakUser> {

	@Autowired
	private KeycloakUserService keycloakUserService;

	@Autowired
	private KeycloakUserPreconditionService preconditionService;

	@Override
	public UpdateControl<OzgKeycloakUser> reconcile(OzgKeycloakUser resource, Context<OzgKeycloakUser> context) {
		try {
			log.info(String.format("Reconciling user %s...", resource.getMetadata().getName()));

			var preconditionError = preconditionService.getReconcilePreconditionErrors(resource);
			if (preconditionError.isPresent()) {
				return buildStatusInProgress(resource, preconditionError.get());
			}

			keycloakUserService.createOrUpdateUser(resource.getSpec(), resource.getMetadata().getNamespace());

			resource.setStatus(buildOzgKeycloakUserOkStatus());
			return UpdateControl.updateStatus(resource);

		} catch (Exception e) {
			log.log(Level.SEVERE, String.format("Could not reconcile user %s for namespace %s: %s", resource.getMetadata().getName(),
					resource.getMetadata().getNamespace(), e.getMessage()), e);

			resource.setStatus(buildOzgKeycloakUserErrorStatus(e.getMessage()));
			return createRescheduleUpdateControl(resource);
		}
	}

	private OzgKeycloakUserStatus buildOzgKeycloakUserOkStatus() {
		return buildOzgKeycloakUserStatus(OzgCustomResourceStatus.OK, null);
	}

	private OzgKeycloakUserStatus buildOzgKeycloakUserErrorStatus(String message) {
		return buildOzgKeycloakUserStatus(OzgCustomResourceStatus.ERROR, message);
	}

	private UpdateControl<OzgKeycloakUser> buildStatusInProgress(OzgKeycloakUser resource, String errorMessage) {
		log.log(Level.WARNING, String.format("Could not reconcile user %s in namespace %s: %s", resource.getMetadata().getName(),
				resource.getMetadata().getNamespace(), errorMessage));

		resource.setStatus(buildOzgKeycloakUserInProgressStatus(errorMessage));
		return createRescheduleUpdateControl(resource);
	}

	private OzgKeycloakUserStatus buildOzgKeycloakUserInProgressStatus(String message) {
		return buildOzgKeycloakUserStatus(OzgCustomResourceStatus.IN_PROGRESS, message);
	}

	private OzgKeycloakUserStatus buildOzgKeycloakUserStatus(OzgCustomResourceStatus status, String message) {
		return OzgKeycloakUserStatus.builder().status(status).message(message).build();
	}

	private UpdateControl<OzgKeycloakUser> createRescheduleUpdateControl(OzgKeycloakUser userResource) {
		return UpdateControl.updateStatus(userResource).rescheduleAfter(Duration.ofSeconds(Config.RECONCILER_RETRY_SECONDS));
	}

	@Override
	public DeleteControl cleanup(OzgKeycloakUser userResource, Context<OzgKeycloakUser> context) {
		try {
			log.info(String.format("Deleting KeycloakUser %s", userResource.getMetadata().getName()));

			keycloakUserService.deleteUser(userResource.getSpec(), userResource.getMetadata().getNamespace());

			return DeleteControl.defaultDelete();
		} catch (Exception e) {
			log.log(Level.SEVERE, String.format("Could not delete user %s in namespace %s", userResource.getMetadata().getName(),
					userResource.getMetadata().getNamespace()), e);
			return DeleteControl.defaultDelete();
		}
	}
}
