/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.operator.keycloak.user;

import java.util.Arrays;
import java.util.Base64;
import java.util.Objects;
import java.util.Optional;
import java.util.logging.Level;

import org.apache.commons.lang3.RandomStringUtils;
import org.apache.commons.lang3.StringUtils;
import org.keycloak.admin.client.CreatedResponseUtil;
import org.keycloak.admin.client.Keycloak;
import org.keycloak.admin.client.resource.RealmResource;
import org.keycloak.representations.idm.ClientRepresentation;
import org.keycloak.representations.idm.RoleRepresentation;
import org.keycloak.representations.idm.UserRepresentation;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import de.ozgcloud.operator.keycloak.KeycloakException;
import de.ozgcloud.operator.keycloak.KeycloakGenericRemoteService;
import de.ozgcloud.operator.keycloak.KeycloakResultParser;
import de.ozgcloud.operator.keycloak.user.OzgKeycloakUserSpec.KeycloakUserSpecUser;
import io.fabric8.kubernetes.api.model.ObjectMeta;
import io.fabric8.kubernetes.api.model.Secret;
import io.fabric8.kubernetes.api.model.SecretBuilder;
import io.fabric8.kubernetes.client.dsl.Resource;
import io.fabric8.kubernetes.client.extension.ResourceAdapter;
import lombok.extern.java.Log;

@Log
@Component
class KeycloakUserRemoteService {

	private static final String SECRET_TYPE = "Opaque";
	static final String SECRET_PASSWORD_FIELD = "password";
	static final String SECRET_NAME_FIELD = "name";

	private static final String USER_NAME_VALIDITY_REGEX = "[^a-zA-Z0-9]";

	@Autowired
	private Keycloak keycloak;
	@Autowired
	private KubernetesRemoteService kubernetesRemoteService;
	@Autowired
	private KeycloakGenericRemoteService keycloakGenericRemoteService;

	public void createUser(UserRepresentation user, String namespace) {
		var realmResource = getRealm(namespace);

		var response = realmResource.users().create(user);
		KeycloakResultParser.parseCreatedResponse(response);

		var userId = CreatedResponseUtil.getCreatedId(response);

		addClientRoles(userId, realmResource, namespace, user);
	}

	public Optional<UserRepresentation> getUserByName(String userName, String namespace) {
		return getRealm(namespace).users().search(userName, true).stream().findFirst();
	}

	public void deleteUser(String userId, String namespace) {
		getRealm(namespace).users().delete(userId);
	}

	public void updateUser(UserRepresentation user, String namespace) {
		var realmResource = getRealm(namespace);

		realmResource.users().get(user.getId()).update(user);

		addClientRoles(user.getId(), realmResource, namespace, user);
	}

	private RealmResource getRealm(String namespace) {
		return keycloak.realm(namespace);
	}

	void addClientRoles(String userId, RealmResource realmResource, String namespace, UserRepresentation user) {
		user.getClientRoles().keySet().forEach(clientId -> {
			var realmClient = getRealmClient(realmResource, clientId);
			user.getClientRoles().get(clientId).stream()
					.map(clientRoleName -> keycloakGenericRemoteService.getClientRole(clientRoleName, realmClient.getId(), namespace)
							.orElseThrow(() -> new KeycloakException(
									"Role " + clientRoleName + " not found for client with clientId " + clientId + " in realm " + namespace)))
					.forEach(clientRole -> addClientRoleToUser(clientRole, realmResource, userId, realmClient));
		});
	}

	ClientRepresentation getRealmClient(RealmResource realmResource, String clientId) {
		return realmResource.clients()
				.findByClientId(clientId).stream().findFirst()
				.orElseThrow(() -> new KeycloakException("Could not find Client with cliendId " + clientId));
	}

	void addClientRoleToUser(RoleRepresentation clientRole, RealmResource realmResource, String userId, ClientRepresentation appClient) {
		realmResource.users().get(userId).roles().clientLevel(appClient.getId()).add(Arrays.asList(clientRole));
	}

	public boolean existSecret(OzgKeycloakUserSpec userSpec, String namespace) {
		return Objects.nonNull(getUserSecret(userSpec, namespace).get());
	}

	public void createSecret(OzgKeycloakUserSpec userSpec, String namespace) {
		log.log(Level.INFO, "Create secret for user: " + userSpec.getKeycloakUser().getUsername());
		var credentialsSecret = createUserSecret(userSpec.getKeycloakUser(), namespace);

		var adapter = new ResourceAdapter<>(getUserSecret(userSpec, namespace));
		adapter.create(credentialsSecret);
	}

	Secret createUserSecret(KeycloakUserSpecUser userSpec, String namespace) {
		return new SecretBuilder()
				.withType(SECRET_TYPE)
				.withMetadata(createMetaData(userSpec, namespace))
				.addToStringData(SECRET_NAME_FIELD, userSpec.getUsername())
				.addToStringData(SECRET_PASSWORD_FIELD, getPassword(userSpec.getPassword()))
				.build();
	}

	private ObjectMeta createMetaData(KeycloakUserSpecUser userSpec, String namespace) {
		var name = buildCredentialSecretName(userSpec);
		var metadata = new ObjectMeta();
		metadata.setName(name);
		metadata.setNamespace(namespace);
		return metadata;
	}

	String getPassword(String userPassword) {
		return StringUtils.isEmpty(userPassword) ? generateRandomPasswordForKeycloak() : userPassword;
	}

	String generateRandomPasswordForKeycloak() {
		log.log(Level.INFO, "Generate password...");
		var upperCaseCharacter = RandomStringUtils.randomAlphabetic(1).toUpperCase();
		var randomString = RandomStringUtils.randomAlphanumeric(7);
		log.log(Level.INFO, "Password generated: " + (upperCaseCharacter + randomString));
		return upperCaseCharacter + randomString;
	}

	public String getPasswordFromSecret(OzgKeycloakUserSpec userSpec, String namespace) {
		return getPasswordFromSecret(getUserSecret(userSpec, namespace));
	}

	public Resource<Secret> getUserSecret(OzgKeycloakUserSpec userSpec, String namespace) {
		var secretName = buildCredentialSecretName(userSpec.getKeycloakUser());
		log.log(Level.INFO, "Get User secret: " + secretName);
		return kubernetesRemoteService.getSecret(namespace, secretName);
	}

	private String buildCredentialSecretName(KeycloakUserSpecUser userSpec) {
		return clarifyName(userSpec.getUsername().toLowerCase()) + "-credentials";
	}

	String clarifyName(String userName) {
		return userName.replaceAll(USER_NAME_VALIDITY_REGEX, StringUtils.EMPTY);
	}

	private String getPasswordFromSecret(Resource<Secret> secret) {
		return decodeBase64(secret.get().getData().get(SECRET_PASSWORD_FIELD));
	}

	private String decodeBase64(String base64String) {
		return new String(Base64.getDecoder().decode(base64String));
	}
}