/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.operator.keycloak.user;

import static org.assertj.core.api.Assertions.*;

import java.util.List;

import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.keycloak.representations.idm.CredentialRepresentation;
import org.mapstruct.factory.Mappers;
import org.mockito.Spy;

import com.fasterxml.jackson.databind.ObjectMapper;

import de.ozgcloud.operator.keycloak.user.OzgKeycloakUserSpec.KeycloakUserSpecClientRole;
import de.ozgcloud.operator.keycloak.user.OzgKeycloakUserSpec.KeycloakUserSpecUser;
import lombok.SneakyThrows;

class KeycloakUserMapperTest {

	@Spy
	private KeycloakUserMapper mapper = Mappers.getMapper(KeycloakUserMapper.class);

	@Nested
	class TestMap {

		@Test
		void shouldMapUsername() {
			var keycloakUser = mapper.map(OzgKeycloakUserSpecTestFactory.create());

			assertThat(keycloakUser.getUsername()).isEqualTo(KeycloakUserSpecUserTestFactory.USERNAME);
		}

		@Test
		void shouldMapFirstname() {
			var keycloakUser = mapper.map(OzgKeycloakUserSpecTestFactory.create());

			assertThat(keycloakUser.getFirstName()).isEqualTo(KeycloakUserSpecUserTestFactory.FIRSTNAME);
		}

		@Test
		void shouldMapLastname() {
			var keycloakUser = mapper.map(OzgKeycloakUserSpecTestFactory.create());

			assertThat(keycloakUser.getLastName()).isEqualTo(KeycloakUserSpecUserTestFactory.LASTNAME);
		}

		@Test
		void userShouldBeEnabled() {
			var keycloakUser = mapper.map(OzgKeycloakUserSpecTestFactory.create());

			assertThat(keycloakUser.isEnabled()).isTrue();
		}

		@Test
		void userShouldBeEmailVerified() {
			var keycloakUser = mapper.map(OzgKeycloakUserSpecTestFactory.create());

			assertThat(keycloakUser.isEmailVerified()).isTrue();
		}

		@Test
		void shouldMapEmail() {
			var keycloakUser = mapper.map(OzgKeycloakUserSpecTestFactory.create());

			assertThat(keycloakUser.getEmail()).isEqualTo(KeycloakUserSpecUserTestFactory.EMAIL);
		}

		@Test
		void shouldSetPasswordType() {
			var keycloakUser = mapper.map(OzgKeycloakUserSpecTestFactory.create());

			assertThat(keycloakUser.getCredentials().get(0).getType()).isEqualTo(CredentialRepresentation.PASSWORD);
		}

		@Disabled("FIXME")
		@Test
		void shouldSetPassword() {
			var keycloakUser = mapper.map(OzgKeycloakUserSpecTestFactory.create());

			assertThat(keycloakUser.getCredentials().get(0).getValue()).isEqualTo(KeycloakUserSpecUserTestFactory.PASSWORD);
		}

		@Test
		void shouldSetPasswordNotTemporary() {
			var keycloakUser = mapper.map(OzgKeycloakUserSpecTestFactory.create());

			assertThat(keycloakUser.getCredentials().get(0).isTemporary()).isFalse();
		}

		@Test
		void shouldMapGroups() {
			var keycloakUser = mapper.map(OzgKeycloakUserSpecTestFactory.create());

			assertThat(keycloakUser.getGroups()).isNotEmpty().contains(KeycloakUserSpecUserTestFactory.GROUP1.getName())
					.contains(KeycloakUserSpecUserTestFactory.GROUP2.getName());
		}

		@Test
		void shouldMapClientRoles() {
			var keycloakUser = mapper.map(OzgKeycloakUserSpecTestFactory.create());

			assertThat(keycloakUser.getClientRoles()).hasSize(1);
		}

		@Test
		void testClientRoleMapper() {
			List<KeycloakUserSpecClientRole> clientRoles = OzgKeycloakUserSpecTestFactory.create().getKeycloakUser().getClientRoles();

			var mappedRoles = mapper.mapClientRoles(clientRoles);

			assertThat(mappedRoles).isNotEmpty()
					.containsKey(KeycloakUserSpecUserTestFactory.CLIENT_NAME)
					.hasSize(1);

			assertThat(mappedRoles.get(KeycloakUserSpecUserTestFactory.CLIENT_NAME))
					.contains(KeycloakUserSpecUserTestFactory.ROLE1.getRoleName(), KeycloakUserSpecUserTestFactory.ROLE2.getRoleName());
		}

		@Test
		void shouldContainClientRoles() {
			var keycloakUser = mapper.map(OzgKeycloakUserSpecTestFactory.create());

			assertThat(keycloakUser.getClientRoles().get(KeycloakUserSpecUserTestFactory.CLIENT_NAME)).hasSize(2)
					.contains(KeycloakUserSpecUserTestFactory.ROLE1.getRoleName(), KeycloakUserSpecUserTestFactory.ROLE2.getRoleName());
		}
	}

	@Nested
	class TestJacksonDeserialization {

		@Test
		@SneakyThrows
		void shouldMapUnsetListToEmptyLists() {
			String json = "{}";

			KeycloakUserSpecUser user = new ObjectMapper().readValue(json, KeycloakUserSpecUser.class);

			assertThat(user.getClientRoles()).isNotNull().isEmpty();
		}

		@Test
		@SneakyThrows
		void shouldMapNullListsToEmptyLists() {
			String json = "{\"client_roles\":null}";

			KeycloakUserSpecUser user = new ObjectMapper().readValue(json, KeycloakUserSpecUser.class);

			assertThat(user.getClientRoles()).isNotNull().isEmpty();
		}
	}

	@Nested
	class TestUpdate {

		private static final String FIRSTNAME2 = "UpdateFirstName";
		private static final String USERID = "UserId";

		@Test
		void shouldMapFirstname() {
			var userRepresentation = UserRepresentationTestFactory.create();
			userRepresentation.setFirstName(FIRSTNAME2);

			var keycloakUser = mapper.update(userRepresentation, OzgKeycloakUserSpecTestFactory.create());

			assertThat(keycloakUser.getFirstName()).isEqualTo(OzgKeycloakUserSpecTestFactory.KEYCLOAK_USER.getFirstName());
		}

		@Test
		void shouldKeepUserId() {
			var userRepresentation = UserRepresentationTestFactory.create();
			userRepresentation.setId(USERID);

			var keycloakUser = mapper.update(userRepresentation, OzgKeycloakUserSpecTestFactory.create());

			assertThat(keycloakUser.getId()).isEqualTo(USERID);
		}
	}
}
