/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.operator.keycloak.user;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Base64;
import java.util.Collections;
import java.util.List;
import java.util.Optional;

import javax.ws.rs.core.Response;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.keycloak.admin.client.Keycloak;
import org.keycloak.admin.client.resource.ClientsResource;
import org.keycloak.admin.client.resource.RealmResource;
import org.keycloak.admin.client.resource.RoleMappingResource;
import org.keycloak.admin.client.resource.RoleScopeResource;
import org.keycloak.admin.client.resource.UserResource;
import org.keycloak.admin.client.resource.UsersResource;
import org.keycloak.representations.idm.ClientRepresentation;
import org.keycloak.representations.idm.RoleRepresentation;
import org.keycloak.representations.idm.UserRepresentation;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.operator.keycloak.KeycloakException;
import de.ozgcloud.operator.keycloak.KeycloakGenericRemoteService;
import de.ozgcloud.operator.keycloak.user.OzgKeycloakUserSpec.KeycloakUserSpecUser;
import io.fabric8.kubernetes.api.model.Secret;
import io.fabric8.kubernetes.api.model.SecretBuilder;
import io.fabric8.kubernetes.client.dsl.Resource;

class KeycloakUserRemoteServiceTest {

	private final static UserRepresentation USER_REPRESENTATION = UserRepresentationTestFactory.create();
	private final static String REALM = "TestRealm";
	private final static String USERID = "UserId";
	private final static String CLIENT_ID = "ClientId";

	private final static String NAMESPACE = "dummyNamespace";

	@Spy
	@InjectMocks
	private KeycloakUserRemoteService userRemoteService;

	@Mock
	private KeycloakGenericRemoteService genericRemoteService;

	@Mock
	private Keycloak keycloak;
	@Mock
	private RealmResource realmResource;
	@Mock
	private UsersResource usersResource;
	@Mock
	private UserResource userResource;
	@Mock
	private ClientsResource clientsResource;
	@Mock
	private ClientRepresentation clientRepresentation;
	@Mock
	private RoleRepresentation roleRepresentation;
	@Mock
	private RoleScopeResource roleScopeResource;
	@Mock
	private RoleMappingResource roleMappingResource;
	@Mock
	private UserRepresentation userRepresentation;
	@Mock
	private KubernetesRemoteService kubernetesRemoteService;

	@Nested
	class TestCreateUser {

		@Mock
		private Response response;

		@BeforeEach
		void init() {
			when(keycloak.realm(REALM)).thenReturn(realmResource);
			when(realmResource.users()).thenReturn(usersResource);
			when(usersResource.create(USER_REPRESENTATION)).thenReturn(response);
			when(response.getStatusInfo()).thenReturn(Response.Status.CREATED);
		}

		@Test
		void shouldThrowOnResponseError() {
			when(response.getStatusInfo()).thenReturn(Response.Status.INTERNAL_SERVER_ERROR);

			assertThrows(KeycloakException.class, () -> userRemoteService.createUser(USER_REPRESENTATION, REALM));
		}

		@Test
		void shouldCallAddClientRoles() {
			when(response.getStatusInfo()).thenReturn(Response.Status.CREATED);
			doNothing().when(userRemoteService).addClientRoles(any(), any(), eq(REALM), any());

			userRemoteService.createUser(USER_REPRESENTATION, REALM);

			verify(userRemoteService).addClientRoles(isNull(), eq(realmResource), eq(REALM), eq(USER_REPRESENTATION));
		}
	}

	@Nested
	class TestGetRealmClient {

		@BeforeEach
		void init() {
			when(realmResource.clients()).thenReturn(clientsResource);
		}

		@Test
		void shouldReturnClientRespresentation() {
			when(clientsResource.findByClientId(UserRepresentationTestFactory.CLIENT_NAME)).thenReturn(List.of(clientRepresentation));

			ClientRepresentation newClientRepresentation = userRemoteService.getRealmClient(realmResource, UserRepresentationTestFactory.CLIENT_NAME);

			assertThat(newClientRepresentation).isSameAs(clientRepresentation);
		}

		@Test
		void shouldThrowOnMissingClient() {
			when(clientsResource.findByClientId(UserRepresentationTestFactory.CLIENT_NAME)).thenReturn(Collections.emptyList());

			assertThrows(KeycloakException.class, () -> userRemoteService.getRealmClient(realmResource, UserRepresentationTestFactory.CLIENT_NAME));
		}
	}

	@Nested
	class TestAddClientRoles {

		@BeforeEach
		void init() {
			when(clientRepresentation.getId()).thenReturn(CLIENT_ID);
			doReturn(clientRepresentation).when(userRemoteService).getRealmClient(any(), any());
		}

		@Test
		void shouldAddClientRole() {
			var user = UserRepresentationTestFactory.create();
			when(genericRemoteService.getClientRole(UserRepresentationTestFactory.ROLE1, CLIENT_ID, REALM))
					.thenReturn(Optional.of(roleRepresentation));
			doNothing().when(userRemoteService).addClientRoleToUser(any(), any(), any(), any());

			userRemoteService.addClientRoles(USERID, realmResource, REALM, user);

			verify(userRemoteService).addClientRoleToUser(roleRepresentation, realmResource, USERID, clientRepresentation);
		}

		@Test
		void shouldThrowOnMissingClientRole() {
			var user = UserRepresentationTestFactory.create();
			userRemoteService.addClientRoles(USERID, realmResource, REALM, userRepresentation);
			when(genericRemoteService.getClientRole(any(), any(), any())).thenReturn(Optional.empty());

			assertThrows(KeycloakException.class,
					() -> userRemoteService.addClientRoles(USERID, realmResource, REALM, user));
		}
	}

	@Nested
	class TestAddClientRoleToUser {

		@BeforeEach
		void init() {
			when(realmResource.users()).thenReturn(usersResource);
			when(usersResource.get(USERID)).thenReturn(userResource);
			when(userResource.roles()).thenReturn(roleMappingResource);
			when(roleMappingResource.clientLevel(CLIENT_ID)).thenReturn(roleScopeResource);

			when(clientRepresentation.getId()).thenReturn(CLIENT_ID);
		}

		@Test
		void shouldAddClientRoleToUser() {
			userRemoteService.addClientRoleToUser(roleRepresentation, realmResource, USERID, clientRepresentation);

			verify(roleScopeResource).add(List.of(roleRepresentation));
		}
	}

	@DisplayName("Update user")
	@Nested
	class TestUpdateUser {

		private final static String USERID = "UserId";

		@BeforeEach
		void init() {
			when(keycloak.realm(REALM)).thenReturn(realmResource);
			when(realmResource.users()).thenReturn(usersResource);
			when(usersResource.get(any())).thenReturn(userResource);

			doNothing().when(userRemoteService).addClientRoles(any(), any(), eq(REALM), any());
		}

		@Test
		void shouldCallKeycloakApiUpdateUser() {
			var user = UserRepresentationTestFactory.create();
			user.setId(USERID);

			userRemoteService.updateUser(user, REALM);

			verify(userResource).update(user);
		}

		@Test
		void shouldCallAddClientRoles() {
			userRemoteService.updateUser(USER_REPRESENTATION, REALM);

			verify(userRemoteService).addClientRoles(isNull(), eq(realmResource), eq(REALM), eq(USER_REPRESENTATION));
		}
	}

	@Nested
	class TestGetUserByName {

		@BeforeEach
		void init() {
			var userList = List.of(UserRepresentationTestFactory.create());

			when(keycloak.realm(REALM)).thenReturn(realmResource);
			when(realmResource.users()).thenReturn(usersResource);
			when(usersResource.search(USER_REPRESENTATION.getUsername(), true)).thenReturn(userList);
		}

		@Test
		void shouldGetUserFromKeycloak() {

			Optional<UserRepresentation> user = userRemoteService.getUserByName(UserRepresentationTestFactory.USERNAME, REALM);

			assertThat(user).isPresent();
		}
	}

	@Nested
	class TestDeleteUser {

		@BeforeEach
		void init() {
			when(keycloak.realm(REALM)).thenReturn(realmResource);
			when(realmResource.users()).thenReturn(usersResource);
		}

		@Test
		void shouldDeleteUser() {
			userRemoteService.deleteUser(USERID, REALM);

			verify(usersResource).delete(USERID);
		}
	}

	@DisplayName("Exists secret")
	@Nested
	class TestExistsSecret {

		private final OzgKeycloakUserSpec userSpec = OzgKeycloakUserSpecTestFactory.create();

		@Mock
		private Resource<Secret> resourceMock;

		@BeforeEach
		void mock() {
			doReturn(resourceMock).when(userRemoteService).getUserSecret(any(), any());
		}

		@Test
		void shouldGetUserSecret() {
			when(resourceMock.get()).thenReturn(new Secret());

			userRemoteService.existSecret(userSpec, NAMESPACE);

			verify(userRemoteService).getUserSecret(userSpec, NAMESPACE);
		}

		@Test
		void shouldReturnTrueIfExists() {
			when(resourceMock.get()).thenReturn(new Secret());

			var exists = userRemoteService.existSecret(userSpec, NAMESPACE);

			assertThat(exists).isTrue();
		}

		@Test
		void shouldReturnFalseIfNotExists() {
			when(resourceMock.get()).thenReturn(null);

			var exists = userRemoteService.existSecret(userSpec, NAMESPACE);

			assertThat(exists).isFalse();
		}
	}

//	@DisplayName("Create Secret")
//	@Nested
//	class TestCreateSecret {
//
//		private final OzgKeycloakUserSpec userSpec = OzgKeycloakUserSpecTestFactory.create();
//
//		@Mock
//		private Resource<Secret> secretResource;
//		@Mock
//		private Secret secret;
//
//		@BeforeEach
//		void mock() {
//			doReturn(secret).when(userRemoteService).createUserSecret(any(), any());
//		}
//
//		@Test
//		void shouldBuildUserSecret() {
//			userRemoteService.createSecret(userSpec, NAMESPACE);
//
//			verify(userRemoteService).createUserSecret(userSpec.getKeycloakUser(), NAMESPACE);
//		}
//
//		@Test
//		void shouldCreateSecretIfNotExists() {
//			userRemoteService.createSecret(userSpec, NAMESPACE);
//
//			verify(server).create(secret);
//		}
//	}

	@DisplayName("Create user secret")
	@Nested
	class TestCreateUserSecret {

		private final KeycloakUserSpecUser userSpec = KeycloakUserSpecUserTestFactory.create();

		@Test
		void shouldHaveType() {
			var secret = userRemoteService.createUserSecret(userSpec, NAMESPACE);

			assertThat(secret.getType()).isEqualTo("Opaque");
		}

		@Test
		void shouldHaveUserName() {
			var secret = userRemoteService.createUserSecret(userSpec, NAMESPACE);

			assertThat(secret.getStringData()).containsEntry(KeycloakUserRemoteService.SECRET_NAME_FIELD, KeycloakUserSpecUserTestFactory.USERNAME);
		}

		@Test
		void shouldHavePassword() {
			doReturn(KeycloakUserSpecUserTestFactory.PASSWORD).when(userRemoteService).getPassword(any());

			var secret = userRemoteService.createUserSecret(userSpec, NAMESPACE);

			assertThat(secret.getStringData()).containsEntry(KeycloakUserRemoteService.SECRET_PASSWORD_FIELD,
					KeycloakUserSpecUserTestFactory.PASSWORD);
		}

		@DisplayName("metadata")
		@Nested
		class TestMetaData {

			@Test
			void shouldHaveName() {
				var secret = userRemoteService.createUserSecret(userSpec, NAMESPACE);

				assertThat(secret.getMetadata().getName()).isEqualTo(userSpec.getUsername() + "-credentials");
			}

			@Test
			void shouldHaveNamespace() {
				var secret = userRemoteService.createUserSecret(userSpec, NAMESPACE);

				assertThat(secret.getMetadata().getNamespace()).isEqualTo(NAMESPACE);
			}
		}
	}

	@DisplayName("Get password")
	@Nested
	class TestGetPassword {

		@Test
		void shouldReturnPasswordIfExists() {
			var password = userRemoteService.getPassword(KeycloakUserSpecUserTestFactory.PASSWORD);

			assertThat(password).isEqualTo(KeycloakUserSpecUserTestFactory.PASSWORD);
		}

		@Test
		void shouldGeneratePasswordIfNotExists() {
			userRemoteService.getPassword(StringUtils.EMPTY);

			verify(userRemoteService).generateRandomPasswordForKeycloak();
		}

		@DisplayName("generate random password for keycloak")
		@Nested
		class TestGenerateRandomPasswordForKeycloak {

			@Test
			void shouldHaveSize() {
				var password = userRemoteService.getPassword(StringUtils.EMPTY);

				assertThat(password).hasSize(8);
			}

			@Test
			void shouldHaveUpperCaseLetterAtFirst() {
				var password = userRemoteService.getPassword(StringUtils.EMPTY);

				assertThat(StringUtils.substring(password, 0, 1)).isUpperCase();
			}

			@Test
			void shouldContainsAlphanumericOnly() {
				var password = userRemoteService.getPassword(StringUtils.EMPTY);

				assertThat(password).isAlphanumeric();
			}
		}
	}

	@DisplayName("Get password from secret")
	@Nested
	class TestGetPasswordFromSecret {

		@Mock
		private Resource<Secret> resource;
		private OzgKeycloakUserSpec userSpec = OzgKeycloakUserSpecTestFactory.create();
		private Secret secret = new SecretBuilder()
				.addToData(KeycloakUserRemoteService.SECRET_PASSWORD_FIELD, Base64.getEncoder().encodeToString("dummyPassword".getBytes()))
				.build();

		@Test
		void shouldGetUserSecret() {
			doReturn(resource).when(userRemoteService).getUserSecret(any(), any());
			when(resource.get()).thenReturn(secret);

			userRemoteService.getPasswordFromSecret(userSpec, NAMESPACE);

			verify(userRemoteService).getUserSecret(userSpec, NAMESPACE);
		}

		@Test
		void shouldReturnDecodedPassword() {
			doReturn(resource).when(userRemoteService).getUserSecret(any(), any());
			when(resource.get()).thenReturn(secret);

			var password = userRemoteService.getPasswordFromSecret(userSpec, NAMESPACE);

			assertThat(password).isEqualTo("dummyPassword");
		}
	}

	@DisplayName("Get user secret")
	@Nested
	class TestGetUserSecret {

		@Test
		void shouldGetSecret() {
			userRemoteService.getUserSecret(OzgKeycloakUserSpecTestFactory.create(), NAMESPACE);

			verify(kubernetesRemoteService).getSecret(NAMESPACE, KeycloakUserSpecUserTestFactory.USERNAME + "-credentials");
		}

		@Test
		void shouldClarifyUserName() {
			userRemoteService.getUserSecret(OzgKeycloakUserSpecTestFactory.create(), NAMESPACE);

			verify(userRemoteService).clarifyName(OzgKeycloakUserSpecTestFactory.KEYCLOAK_USER.getUsername().toLowerCase());
		}
	}

	@DisplayName("Clarify name")
	@Nested
	class TestClarifyName {

		@ValueSource(strings = { "_user_name_", ".user.name.", "-user-name-" })
		@ParameterizedTest
		void shouldReplaceForbiddenCharacter(String userName) {
			var clarifiedName = userRemoteService.clarifyName(userName);

			assertThat(clarifiedName).isEqualTo("username");
		}
	}
}