/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.operator.keycloak.user;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.keycloak.representations.idm.UserRepresentation;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.Spy;

import io.fabric8.kubernetes.api.model.Secret;
import io.fabric8.kubernetes.client.dsl.Resource;

class KeycloakUserServiceTest {

	private static final String TEST_NAMESPACE = "TestNamespace";
	private static final String TEST_USERID = "TestUserId";
	private static final String PASSWORD = "CreatedDummyPassword";

	@Spy
	@InjectMocks
	private KeycloakUserService service;

	@Mock
	private KeycloakUserRemoteService remoteService;

	@Mock
	private KeycloakUserMapper userMapper;

	@DisplayName("Create or Update")
	@Nested
	class TestCreateOrUpdate {

		@Mock
		private Resource<Secret> resource;
		@Captor
		private ArgumentCaptor<OzgKeycloakUserSpec> ozgKeycloakUserSpecCaptor;

		private final OzgKeycloakUserSpec userSpec = OzgKeycloakUserSpecTestFactory.create();

		@BeforeEach
		void mock() {
			when(remoteService.getPasswordFromSecret(any(), any())).thenReturn(PASSWORD);
		}

		@Test
		void shouldVerifiySecretExists() {
			service.createOrUpdateUser(userSpec, TEST_NAMESPACE);

			verify(remoteService).existSecret(userSpec, TEST_NAMESPACE);
		}

		@Test
		void shouldCreateSecretIfNotExists() {
			when(remoteService.existSecret(any(), any())).thenReturn(false);

			service.createOrUpdateUser(userSpec, TEST_NAMESPACE);

			verify(remoteService).createSecret(userSpec, TEST_NAMESPACE);
		}

		@Test
		void shouldGetPasswordFromSecret() {
			service.createOrUpdateUser(userSpec, TEST_NAMESPACE);

			verify(remoteService).getPasswordFromSecret(userSpec, TEST_NAMESPACE);
		}

		@Test
		void shouldSetPasswortFromSecret() {
			var userWithoutPassword = OzgKeycloakUserSpecTestFactory.createBuilder()
					.keycloakUser(KeycloakUserSpecUserTestFactory.createBuiler().password(StringUtils.EMPTY).build()).build();

			service.createOrUpdateUser(userWithoutPassword, TEST_NAMESPACE);

			verify(userMapper).map(ozgKeycloakUserSpecCaptor.capture());
			assertThat(ozgKeycloakUserSpecCaptor.getValue().getKeycloakUser().getPassword()).isEqualTo(PASSWORD);
		}

		@Test
		void shouldCallUserMapper() {
			service.createOrUpdateUser(userSpec, TEST_NAMESPACE);

			verify(userMapper).map(userSpec);
		}

		@Test
		void shouldCallUserRemoteServiceGetUserByName() {
			var userRepresentation = UserRepresentationTestFactory.create();

			service.createOrUpdateUser(userSpec, TEST_NAMESPACE);

			verify(remoteService).getUserByName(userRepresentation.getUsername(), TEST_NAMESPACE);
		}

		@Test
		void shouldCreateUserIfNotExists() {
			when(remoteService.getUserByName(OzgKeycloakUserSpecTestFactory.KEYCLOAK_USER.getUsername(), TEST_NAMESPACE))
					.thenReturn(Optional.empty());
			when(userMapper.map(any())).thenReturn(Mockito.mock(UserRepresentation.class));

			service.createOrUpdateUser(OzgKeycloakUserSpecTestFactory.create(), TEST_NAMESPACE);

			verify(remoteService).createUser(any(UserRepresentation.class), eq(TEST_NAMESPACE));
		}

		@Test
		void shouldUpdateUserIfExists() {
			var userRepresentation = Mockito.mock(UserRepresentation.class);
			when(remoteService.getUserByName(OzgKeycloakUserSpecTestFactory.KEYCLOAK_USER.getUsername(), TEST_NAMESPACE))
					.thenReturn(Optional.of(userRepresentation));
			when(userMapper.update(eq(userRepresentation), any())).thenReturn(userRepresentation);

			service.createOrUpdateUser(OzgKeycloakUserSpecTestFactory.create(), TEST_NAMESPACE);

			verify(remoteService).updateUser(userRepresentation, TEST_NAMESPACE);
		}
	}

	@Nested
	class TestDeleteUser {

		@Test
		void shouldCallUserMapper() {

			var testUser = OzgKeycloakUserSpecTestFactory.create();

			service.deleteUser(testUser, TEST_NAMESPACE);

			verify(userMapper).map(testUser);
		}

		@Test
		void shouldCallUserRemoteServiceGetUser() {

			var userRepresentation = UserRepresentationTestFactory.create();
			when(userMapper.map(any())).thenReturn(userRepresentation);

			service.deleteUser(OzgKeycloakUserSpecTestFactory.create(), TEST_NAMESPACE);

			verify(remoteService).getUserByName(userRepresentation.getUsername(), TEST_NAMESPACE);
		}

		@Test
		void shouldCallUserRemoteServiceDeleteUser() {

			var userRepresentation = UserRepresentationTestFactory.create();
			when(userMapper.map(any())).thenReturn(userRepresentation);
			when(remoteService.getUserByName(UserRepresentationTestFactory.USERNAME, TEST_NAMESPACE)).thenReturn(Optional.of(userRepresentation));
			userRepresentation.setId(TEST_USERID);

			service.deleteUser(OzgKeycloakUserSpecTestFactory.create(), TEST_NAMESPACE);

			verify(remoteService).deleteUser(TEST_USERID, TEST_NAMESPACE);
		}
	}
}
